--------------------------------------------------
-- Standard Evaluierer
--------------------------------------------------

data Term = Con Int | 
	    Div Term Term

eval :: Term -> Int
eval (Con x) = x
eval (Div t u) = (eval t) `div` (eval u)

answer,wrong :: Term
answer = (Div (Div (Con 1972) (Con 2)) (Con 23))
wrong = (Div (Con 2) (Div (Con 1) (Con 0)))

--------------------------------------------------
-- Exception Evaluierer
--------------------------------------------------

data Exc a = Raise Exception | Return a
type Exception = String

evalExc :: Term -> Exc Int
evalExc (Con x) = Return x
evalExc (Div t u) 
    = h (evalExc t)
      where 
      h (Raise e) = Raise e
      h (Return x) = h'(evalExc u)
		     where 
		     h' (Raise e) = Raise e
		     h' (Return y) 
			 = if y == 0
			   then Raise "Division by zero"
			   else Return (x `div` y)

instance Show a =>Show (Exc a) where
    show (Raise e) = "exception: " ++ show e 
    show (Return x) = "Ergebnis: " ++ show x

--------------------------------------------------
-- State Evaluierer
--------------------------------------------------
newtype St a = MkSt(State -> (a,State))
type State = Int

apply :: St a -> State -> (a,State)
apply (MkSt f) s = f s

evalSt :: Term -> St Int
evalSt (Con x) = MkSt f
	       where f s = (x,s)
evalSt (Div t u) = MkSt f 
		 where 
		 f s = (x `div` y, s'' +1)
		       where
		       (x,s') = apply (evalSt t) s
		       (y,s'') = apply (evalSt u) s'

instance Show a => Show (St a) where
    show f = "value: " ++ show x ++ "\ncount: " ++ show s
	     where (x,s) = apply f 0

--------------------------------------------------
-- Trace Output Evaluierer
--------------------------------------------------

newtype Out a = MkOut(Output,a)
type Output = String

evalOut :: Term -> Out Int
evalOut (Con x) = MkOut(line (Con x) x,x)
evalOut (Div t u) = MkOut(ox ++ oy ++ line (Div t u) z,z)
 		    where MkOut (ox,x) = evalOut t
 			  MkOut (oy,y) = evalOut u
			  z = x `div` y

line :: Term -> Int -> Output
line t x = "Term: " ++  show t ++ " ergibt: " ++  show x ++ "\n"

instance Show a => Show(Out a) where
    show (MkOut(ox,x)) = ox ++ "Ergebnis: " ++  show x

instance Show (Term) where
    show (Con x) = "(Con " ++ show x ++ ")"
    show (Div t u) = "(Div " ++ show t ++ " " ++ show u ++ ")"
