--------------------------------------------------
-- Standard Evaluierer mit Monad
--------------------------------------------------

data Term = Con Int | Div Term Term

eval :: Monad m => Term -> m Int
eval (Con x) = return x
eval (Div t u) = do x <- eval t
		    y <- eval u
		    return (x `div` y)

answer,wrong :: Term
answer = (Div (Div (Con 1972) (Con 2)) (Con 23))
wrong = (Div (Con 2) (Div (Con 1) (Con 0)))

--------------------------------------------------
-- Standard Evaluierer mit ID-Monad
--------------------------------------------------

newtype Id a = MkId a

instance Monad Id where
    return x = MkId x
    (MkId x) >>= q = q x

instance Show a => Show (Id a) where
    show (MkId x) = "Ergebnis: " ++ show x

evalId :: Term -> Id Int
evalId = eval

--------------------------------------------------
-- Exception Evaluierer mit Monad
--------------------------------------------------

data Exc a = Raise Exception | Return a
type Exception = String

instance Monad Exc where
    return x = Return x
    (Raise e) >>= q = Raise e
    (Return x) >>= q = q x

raise :: Exception -> Exc a
raise e = Raise e

evalExc :: Term -> Exc Int
evalExc (Con x) = return x
evalExc (Div t u) = do x <- evalExc t
		       y <- evalExc u
		       if y == 0
			  then raise "division by zero"
			  else return (x `div` y)

instance Show a =>Show (Exc a) where
    show (Raise e) = "exception: " ++ show e 
    show (Return x) = "Ergebnis: " ++ show x

--------------------------------------------------
-- State Evaluierer mit Monad
--------------------------------------------------

newtype St a = MkSt(State -> (a,State))
type State = Int

apply :: St a -> State -> (a,State)
apply (MkSt f) s = f s

instance Monad St where 
    return x = MkSt f where f s = (x,s)
    p >>= q = MkSt f
	      where 
	      f s = apply (q x) s'
		    where (x,s') = apply p s

tick :: St()
tick = MkSt f where f s = ((),s+1)

evalSt :: Term -> St Int
evalSt (Con x) = return x
evalSt (Div t u) = do x <- evalSt t
		      y <- evalSt u
		      tick 
		      return (x `div` y)

instance Show a => Show (St a) where
    show f = "value: " ++ show x ++ "\ncount: " ++ show s
	     where (x,s) = apply f 0

--------------------------------------------------
-- Output Evaluierer mit Monad
--------------------------------------------------

newtype Out a = MkOut(Output,a)
type Output = String

instance Monad Out where
    return x = MkOut("",x)
    p >>= q = MkOut (ox ++ oy,y)
	      where MkOut (ox,x) = p
		    MkOut (oy,y) = q x

out :: Output -> Out()
out ox = MkOut (ox,())

evalOut :: Term -> Out Int
evalOut (Con x) = do out(line(Con x)x)
		     return x
evalOut (Div t u) = do x <- evalOut t
		       y <- evalOut u
		       out(line(Div t u) (x `div` y))
		       return (x `div` y)

line :: Term -> Int -> Output
line t x = "Term: " ++  show t ++ " ergibt: " ++  show x ++ "\n"

instance Show a => Show(Out a) where
    show (MkOut(ox,x)) = ox ++ "Ergebnis: " ++  show x

instance Show (Term) where
    show (Con x) = "(Con " ++ show x ++ ")"
    show (Div t u) = "(Div " ++ show t ++ " " ++ show u ++ ")"
